from typing import List, Dict
from uuid import uuid4

from hathor.serve.models import ChangeListEntry, Package, PackageTakeout

CLIENTS = dict()
THREAD_INTERVAL = 2


def by_id(client_id: str):
    return CLIENTS.get(client_id, None)


class OnClientAdded:
    _listener: callable
    _listeners: List[callable] = []

    def __init__(self, listener: callable):
        self._listener = listener
        self._listeners.append(listener)

    def disconnect(self):
        self._listeners.remove(self._listener)

    @classmethod
    def fire(cls, client):
        for listener in cls._listeners:
            listener(client)


class OnClientRemoved:
    _listener: callable
    _listeners: List[callable] = []

    def __init__(self, listener: callable):
        self._listener = listener
        self._listeners.append(listener)

    def disconnect(self):
        self._listeners.remove(self._listener)

    @classmethod
    def fire(cls, client):
        for listener in cls._listeners:
            listener(client)


def _generate_client_id() -> str:
    return str(uuid4())


class Client:
    _id: str
    _name: str
    _changes: Dict[str, ChangeListEntry]
    _packages: Dict[str, Package]

    def __init__(self, name):
        self._id = _generate_client_id()
        self._name = name

        self._changes = dict()
        self._packages = dict()

        CLIENTS[self._id] = self
        OnClientAdded.fire(self)

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self):
        return self._name

    def add_changes(self, changes: List[ChangeListEntry]):
        for change in changes:
            self._changes[change.id] = change

    def remove(self):
        CLIENTS.pop(self._id)
        OnClientRemoved.fire(self)

    @property
    def changes(self) -> List[ChangeListEntry]:
        return list(self._changes.values())

    def delete_changes(self, changes: List[str]):
        for change in changes:
            if change in self._changes:
                self._changes.pop(change)

    def create_package(self, change_ids: List[str]) -> Package:
        package = Package(
            id=str(uuid4()),
            changes=list(
                map(lambda t: t[1],
                    filter(lambda t: t[0] in change_ids,
                           self._changes.items()
                           )
                    )
            )
        )

        self._packages[package.id] = package

        return package

    def create_takeout(self, package_id: str) -> PackageTakeout:
        package = self._packages[package_id]

        takeout = PackageTakeout(dict())

        for change in package.changes:
            if not takeout.add_change(change):
                break

            if change.id in self._changes:
                self._changes.pop(change.id)

        return takeout

    def delete_package(self, package_id):
        if package_id in self._packages:
            self._packages.pop(package_id)


def iter_clients():
    for client in CLIENTS.values():
        yield client



import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import List, Union

from fs.base import FS

from hathor import resources

ENTRY_POINT_PTN = re.compile(r"---\s*@entrypoint\s+(\w+)\s+([\w\d.]+)")
REQUIRE_PTN = re.compile(r"require\([\"\']([\w\d\.]+)[\'\"]\)")
QUOTES = "'" + '"'


class EntryPointType:
    script = "Script"
    local_script = "LocalScript"


@dataclass
class EntryPointDefinition:
    line: str
    type: str
    game_location: str


@dataclass
class RequireDefinition:
    line: str
    require: str

    def __str__(self):
        return self.require


def module_returns(lines: List[str]) -> bool:
    for line in lines[-3:]:
        if line.strip().startswith("return"):
            return True

    return False


class LuaFile:
    _entry_point: Union[None, EntryPointDefinition] = None
    _requires: List[RequireDefinition]
    _root_fs: FS
    _path: str

    def __init__(self, root_fs: FS, path: str):
        self._root_fs = root_fs
        self._path = path
        self._requires = []
        self._scan()

    def _scan(self):
        with self._root_fs.open(self._path) as fp:
            for line in fp.readlines():
                if match := ENTRY_POINT_PTN.match(line):
                    self._entry_point = EntryPointDefinition(
                        line,
                        match.group(1).strip(),
                        match.group(2).strip()
                    )

                if match := REQUIRE_PTN.search(line.strip()):
                    self._requires.append(RequireDefinition(
                        line,
                        match.group(1).strip()
                    ))

    @property
    def is_entry_point(self):
        return self._entry_point is not None

    @property
    def require_alias(self) -> str:
        path_split = self._path.lstrip(os.path.sep).split(os.path.sep)
        last = len(path_split) - 1
        file_name_split = path_split[last].split(".")
        path_split[last] = ".".join(file_name_split[:-1])

        return ".".join(path_split)

    @property
    def requires(self) -> List[RequireDefinition]:
        return list(self._requires)

    @property
    def name(self):
        return ".".join(os.path.basename(self._path).split('.')[:-1])

    @property
    def entry_point_path(self) -> Path:
        assert self.is_entry_point, "File must be an entry point"

        return Path(*self._entry_point.game_location.split("."))

    @property
    def path(self) -> str:
        return self._path

    def read(self):
        with self._root_fs.open(self._path, "r") as fp:
            contents = fp.read()

        return contents

    def processed_contents(self):
        contents = self.read()
        contents = resources.hathor_stub() + "\n" + contents

        processed_lines = []

        for line in contents.splitlines():
            if match := REQUIRE_PTN.search(line):
                proc_line = REQUIRE_PTN.sub(f"__require(\"{match.group(1)}\")", line)
                processed_lines.append(proc_line)

                continue

            processed_lines.append(line)

        if self.is_entry_point and not module_returns(processed_lines):
            processed_lines.append("\n")
            processed_lines.append("return true")

        contents = "\n".join(processed_lines)

        return contents

    def __str__(self):
        return self.require_alias



from fs.base import FS

from hathor import resources
from hathor.project.information.project import Project
from hathor.project.information.requirement_explorer import RequirementExplorer


def build_modular(project: Project, build_root: FS):
    build_root.makedirs("/", recreate=True)

    for entry_point in project.entry_points:
        epp = build_root.makedirs(str(entry_point.entry_point_path.parent), recreate=True)

        modules_directory_name = entry_point.name + "_Modules"
        modules_directory = epp.makedirs(modules_directory_name, recreate=True)

        explorer = RequirementExplorer(project, entry_point)

        def emit_modules():
            to_emit = {
                entry_point.require_alias: entry_point
            }

            for link in explorer.links:
                for requirement in link.requirements:
                    to_emit[requirement.require_alias] = requirement

            for require_str, file in to_emit.items():
                with modules_directory.open(require_str + ".lua", "w+") as fp:
                    fp.write(file.processed_contents())

        def emit_loader():
            with modules_directory.open("__HathorLoader.lua", "w+") as fp:
                fp.write(resources.hathor_loader())

        def emit_runner():
            with epp.open(f"{entry_point.name}.Run.lua", "w+") as fp:
                fp.write(resources.hathor_run(modules_directory_name, entry_point.require_alias))

        emit_modules()
        emit_loader()
        emit_runner()



import hashlib
import logging
import os
import time
import uuid
from threading import Thread
from typing import Dict, Callable, List

from fs.base import FS
from fs.info import Info
from fs.memoryfs import MemoryFS

from hathor.project.builders import Builder, get_builder
from hathor.project.information.project import Project
from hathor.serve import change_tracker
from hathor.serve.client import iter_clients, Client, OnClientAdded
from hathor.serve.models import ChangeListEntry, Kind, KindType

LOG = logging.getLogger(__name__)

ChangeList = Dict[str, ChangeListEntry]


def _tuple_path(p: str):
    return tuple(list(filter(lambda s: not not s.strip(), p.split(os.path.sep))))


def _append_change(change_list: ChangeList, build_root: FS, path: str, info: Info, kind: KindType):
    change_list[path] = ChangeListEntry(
        id=str(uuid.uuid4()),
        kind=kind,
        path=_tuple_path(path),
        isDirectory=info.is_dir,
        isFile=info.is_file,
        fs=build_root
    )


def _hash_string(s: str):
    h = hashlib.blake2b()
    h.update(s.encode("UTF-8"))

    digest = h.hexdigest()

    return digest.lower()


def _diff_hash_table(left, right) -> List[str]:
    diff = set()

    def go(left_table, right_table):
        for left_key, left_value in left_table.items():
            right_value = right_table.get(left_key)
            if left_value != right_value:
                diff.add(left_key)

    go(left, right)
    go(right, left)

    return list(diff)


class ProjectBuilderThread(Thread):
    _running: bool = True
    _project: Project
    _builder: Builder
    _interval: float

    _build_root: MemoryFS = None
    _hash_table: Dict[str, str] = None

    _name: str
    _log: logging.Logger

    _on_client_added_handle: OnClientAdded

    def __init__(self, project: Project, builder: Builder, interval: float):
        super().__init__()

        self._project = project
        self._builder = builder
        self._interval = float(interval)

        self._build_root = MemoryFS()
        self._hash_table = dict()

        self._name = f"Project builder for '{project.name}'"
        self.setName(self._name)

        self._log = logging.getLogger(self._name)

        self._on_client_added_handle = OnClientAdded(self._on_client_added)

    def _build(self):
        self._log.debug("Going to build the project")

        t0 = time.time()

        self._project.rescan()
        self._builder(self._project, self._build_root)

        dt = time.time() - t0
        LOG.debug(f"Building the project took {round(dt, 5)} seconds")

    def _iter_files(self, f: Callable[[str, Info], None]):
        if self._build_root:
            for path, info in self._build_root.glob("**/*"):
                if not self._build_root.isfile(path):
                    continue

                f(path, info)

    def _update_hash_table(self):
        self._hash_table = dict()

        def hash_file(path: str, info: Info):
            with self._build_root.open(path, "r") as fp:
                self._hash_table[path] = _hash_string(fp.read())

        self._iter_files(hash_file)

        return dict(self._hash_table)

    def _on_client_added(self, client: Client):
        change_list: ChangeList = dict()

        self._iter_files(lambda path, info: _append_change(
            change_list=change_list,
            build_root=self._build_root,
            path=path,
            info=info,
            kind=Kind.create
        ))

        client.add_changes(list(change_list.values()))

    def _prepare_changes(self) -> ChangeList:
        change_list: ChangeList = dict()

        self._iter_files(lambda path, info: _append_change(
            change_list=change_list,
            build_root=self._build_root,
            path=path,
            info=info,
            kind=Kind.delete
        ))

        return change_list

    def _distribute_changes(self, change_list: ChangeList):
        for client in iter_clients():
            client.add_changes(list(change_list.values()))

    def _clean(self):
        for f in self._build_root.listdir("/"):
            if self._build_root.isdir(f):
                self._build_root.removetree(f)

            elif self._build_root.isfile(f):
                self._build_root.remove(f)

    def run(self) -> None:
        self._log.debug("Running the project builder")

        def _wait(seconds: float, interval: float = 0.1):
            while seconds > 0 and self._running:
                time.sleep(interval)
                seconds -= interval

        while self._running:
            changes = change_tracker.take_changes()

            if len(changes) > 0:
                # Setup
                table_before = self._update_hash_table()
                change_list = self._prepare_changes()

                # Clean build in memory
                self._clean()
                self._build()

                # Undo all faux removals
                for path, change in dict(change_list).items():
                    if self._build_root.exists(path) and change.kind == Kind.delete:
                        change_list.pop(path)

                # Add all files from the memory fs
                self._iter_files(lambda p, info: _append_change(
                    change_list=change_list,
                    build_root=self._build_root,
                    path=p,
                    info=info,
                    kind=Kind.modify
                ))

                # Find all real changes
                table_after = self._update_hash_table()
                hash_table_changes = _diff_hash_table(table_before, table_after)

                # Filter out unchanged files using the hash table
                reduced_changelist: ChangeList = dict()
                for path, change in change_list.items():
                    if path in hash_table_changes:
                        reduced_changelist[path] = change

                # Distribute all changes
                self._distribute_changes(reduced_changelist)

            try:
                _wait(self._interval)

            except Exception as e:
                self._running = False
                self._log.info(f"time.sleep interrupt: {str(e)}")

    def stop(self):
        self._log.info("Stopping project builder")
        self._running = False

    def __del__(self):
        self._on_client_added_handle.disconnect()


class State:
    thread: ProjectBuilderThread = None


def stop():
    if State.thread:
        LOG.debug("at_exit handler is going to stop a project builder thread")
        State.thread.stop()
        LOG.debug("Joining that thread now")
        State.thread.join()


def configure(project: Project, builder: str, interval: int):
    assert State.thread is None, "The project builder is already configured"

    State.thread = ProjectBuilderThread(
        project,
        get_builder(builder),
        interval
    )

    State.thread.start()



import logging
from pathlib import Path

from fs.base import FS
from watchdog.events import FileSystemEventHandler, FileCreatedEvent, FileDeletedEvent, FileModifiedEvent
from watchdog.observers import Observer

from hathor.project.information.project import Project
from hathor.serve import change_tracker
from hathor.serve.change_tracker import FSChange
from hathor.serve.models import KindType, Kind

LOG = logging.getLogger(__name__)


class ChangeHandler(FileSystemEventHandler):
    _fs: FS
    _root: Path
    _project: Project

    def __init__(self, project: Project, fs: FS):
        self._project = project
        self._fs = fs
        self._root = Path(fs.root_path)

    def create_change_entry(self, kind: KindType, src_path: str):
        real_path = Path(src_path).absolute()
        relative_path = real_path.relative_to(self._root)

        change = FSChange(
            project=self._project,
            fs=self._fs,
            kind=kind,
            real_path=real_path,
            relative_path=relative_path
        )

        LOG.debug(change)

        change_tracker.notify(change)

    def on_created(self, event):
        if isinstance(event, FileCreatedEvent):
            self.create_change_entry(Kind.create, event.src_path)

    def on_deleted(self, event):
        if isinstance(event, FileDeletedEvent):
            self.create_change_entry(Kind.delete, event.src_path)

    def on_modified(self, event):
        if isinstance(event, FileModifiedEvent):
            self.create_change_entry(Kind.modify, event.src_path)


def watch_files(project: Project):
    observer = Observer()

    for fs in project.source_file_systems:
        fs: FS = fs

        handler = ChangeHandler(project, fs)
        root_path = fs.root_path

        for path, info in fs.glob("**/*.lua"):
            absolute_path = Path(root_path, path.lstrip("/"))
            handler.create_change_entry(Kind.modify, absolute_path)

        observer.schedule(handler, root_path, recursive=True)

    observer.start()



from typing import List, Dict

from fs.base import FS

from hathor.project.information.lua_file import LuaFile


class SourceRoot:
    _fs: FS
    _properties: Dict
    _lua_files: List[LuaFile]

    def __init__(self, fs: FS, properties: Dict):
        self._fs = fs
        self._properties = properties
        self._lua_files = []

        self._scan()

    def _scan(self):
        self._lua_files.extend(map(lambda gm: LuaFile(self._fs, gm.path), self._fs.glob("**/*.lua")))

    @property
    def lua_files(self) -> List[LuaFile]:
        return list(self._lua_files)

    @property
    def path(self):
        return self._path

    @property
    def fs(self):
        return self._fs



import os
from dataclasses import dataclass
from typing import Tuple, List, Dict

from fs.base import FS

KindType = str


@dataclass()
class Kind:
    create: KindType = "ENTRY_CREATE"
    delete: KindType = "ENTRY_DELETE"
    modify: KindType = "ENTRY_MODIFY"


@dataclass()
class ChangeListEntry:
    id: str
    kind: KindType
    path: Tuple[str]
    isDirectory: bool
    isFile: bool
    fs: FS


@dataclass()
class Package:
    id: str
    changes: List[ChangeListEntry]


ONE_MEG = 1024 * 1024


@dataclass()
class PackageTakeout:
    content: Dict[str, str]
    remainingChanges: int = 0
    packageSize: int = 0
    content_size: int = 0
    size_limit = 4 * ONE_MEG

    def add_change(self, change: ChangeListEntry) -> bool:
        if change.kind != Kind.delete:
            with change.fs.open(os.path.sep.join(change.path), "r") as fp:
                data = fp.read()

        else:
            data = ""

        if len(data) + self.packageSize > self.size_limit:
            return False

        self.content[change.id] = data
        self.content_size += len(data)

        self.packageSize = len(self.content)

        return True



from dataclasses import dataclass
from typing import Dict, List

from hathor.project.information.lua_file import LuaFile
from hathor.project.information.project import Project


@dataclass
class RequireLink:
    file: LuaFile
    requirements: List[LuaFile]


class RequirementExplorer:
    _project: Project
    _entry_point: LuaFile
    _links: List[RequireLink]

    def __init__(self, project: Project, entry_point: LuaFile):
        self._project = project
        self._entry_point = entry_point
        self._links = []

        self._build()

    def _build(self):
        reference: Dict[str, LuaFile] = \
            dict(zip(map(lambda f: f.require_alias, self._project.lua_files), self._project.lua_files))

        def _process_file(file: LuaFile):
            requirements: List[LuaFile] = []

            for require_def in file.requires:
                if required_file := reference.get(require_def.require, None):
                    requirements.append(required_file)

                else:
                    raise RuntimeError(str(require_def))

            self._links.append(RequireLink(file, requirements))

            for requirement in requirements:
                _process_file(requirement)

        _process_file(self._entry_point)

    @property
    def links(self) -> List[RequireLink]:
        return list(self._links)



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

from fs.base import FS

from hathor.project.information.project import Project
from hathor.serve.models import Kind, KindType

LOG = logging.getLogger(__name__)


@dataclass()
class FSChange:
    project: Project
    fs: FS
    kind: KindType
    real_path: Path
    relative_path: Path

    @property
    def path_string(self):
        return "/".join(self.relative_path.parts)


Changes = Dict[str, FSChange]


class State:
    changes: Changes = {}


def notify(change: FSChange):
    changes = State.changes
    path = change.path_string

    if change.kind == Kind.create or change.kind == Kind.modify:
        changes[path] = change

    elif change.kind == Kind.delete and path in changes:
        changes.pop(path)

    LOG.debug("--- CHANGES ---")
    for change in changes.values():
        LOG.debug((change.path_string, change.kind))


def take_changes():
    changes = State.changes
    State.changes = dict()

    return changes



import dataclasses
import logging
import os
from typing import Tuple

from flask import Flask, request, abort
from flask_restx import Api, fields, Resource

import hathor.serve.client as hathor_client
from hathor.project.information.project import Project
from hathor.serve import project_builder
from hathor.serve.client import Client
from hathor.serve.models import ChangeListEntry
from hathor.serve.watch_files import watch_files

LOG = logging.getLogger(__name__)

app = Flask("Hathor")

api = Api(
    app,
    version="1.0",
    title="Hathor Synchronization API",
    authorizations={
        "client_id": {
            "type": "apiKey",
            "in": "header",
            "name": "Authorization"
        }
    }
)


class Namespaces:
    announce = api.namespace("announce", "Client announcements")
    changelist = api.namespace("changelist", "Changelist API")
    package = api.namespace("package", "Synchronization Packages")


class Models:
    announcement = api.model("Announcement", {
        "name": fields.String(description="The name for the new client")
    })

    announcement_response = api.model("AnnouncementResponse", {
        "sessionId": fields.String(description="The session ID for the announced client")
    })

    changelist_entry = api.model("ChangeListEntry", {
        "id": fields.String(description="The unique ID for this change"),
        "kind": fields.String(description="The kind of change"),
        "path": fields.List(fields.String()),
        "isDirectory": fields.Boolean(),
        "isFile": fields.Boolean()
    })

    change_delete_request = api.model("ChangeDeleteRequest", {
        "changes": fields.List(fields.String())
    })

    package_takeout = api.model("PackageTakeout", {
        "content": fields.String(),
        "remainingChanges": fields.Integer(),
        "packageSize": fields.Integer()
    })

    create_package_request = api.model("CreatePackageRequest", {
        "changes": fields.List(fields.String())
    })

    package = api.model("Package", {
        "id": fields.String()
    })

    delete_package_request = api.model("DeletePackageRequest", {
        "packageId": fields.String()
    })


def auth_token():
    header = request.headers.get("Authorization")
    if header is None:
        return

    header_spl = header.strip().split(" ")
    assert header_spl[0].strip().lower() == "bearer"

    return header_spl[1]


def find_client() -> Client:
    if client_id := auth_token():
        if client := hathor_client.by_id(client_id):
            return client

    abort(403)


def doc_auth_token(ns):
    def deco(*args):
        return ns.param("Authorization", "Client ID", _in="header")(*args)

    return deco


@Namespaces.announce.route("")
class Announcements(Resource):
    @Namespaces.announce.expect(Models.announcement)
    @Namespaces.announce.marshal_with(Models.announcement_response)
    def post(self):
        data = request.json
        client = Client(data["name"])

        return {
            "sessionId": client.id
        }

    @api.doc(security="client_id")
    @doc_auth_token(Namespaces.announce)
    def delete(self):
        client = find_client()
        client.remove()

        return "OK", 200


@Namespaces.changelist.route("")
class ChangeList(Resource):
    @doc_auth_token(Namespaces.changelist)
    @Namespaces.changelist.marshal_list_with(Models.changelist_entry)
    def get(self):
        client = find_client()

        def change_to_dict(change: ChangeListEntry):
            return {
                "id": change.id,
                "kind": change.kind,
                "path": change.path,
                "isDirectory": change.isDirectory,
                "isFile": change.isFile
            }

        return list(map(change_to_dict, client.changes))

    @doc_auth_token(Namespaces.changelist)
    @Namespaces.changelist.expect(Models.change_delete_request)
    def delete(self):
        client = find_client()
        client.delete_changes(request.json["changes"])

        return "OK"


@Namespaces.package.route("")
class Packages(Resource):
    @Namespaces.package.param("id", description="The package ID to retrieve")
    @doc_auth_token(Namespaces.package)
    # @Namespaces.package.marshal_with(Models.package_takeout)
    def get(self):
        client = find_client()

        takeout = dataclasses.asdict(client.create_takeout(request.args["id"]))

        return takeout

    @doc_auth_token(Namespaces.package)
    @Namespaces.package.expect(Models.create_package_request)
    @Namespaces.package.marshal_with(Models.package)
    def post(self):
        client = find_client()
        package = client.create_package(request.json["changes"])

        return {
            "id": package.id
        }

    @doc_auth_token(Namespaces.package)
    @Namespaces.package.expect(Models.delete_package_request)
    def delete(self):
        client = find_client()
        client.delete_package(request.json["packageId"])


def get_tcp_bind(project: Project) -> Tuple[str, int]:
    host = project.serve_config.get("host", "127.0.0.1")
    port = str(project.serve_config.get("port", 8080))
    assert port.isnumeric(), f"The provided port {port} is not numerical"
    port = int(port)

    return host, port


def run(project: Project):
    host, port = get_tcp_bind(project)
    debug = int(os.environ.get("DEBUG", "0").strip()) > 0

    watch_files(project)

    preferred_profile = project.serve_config.get("build_profile", "default")
    project.active_profile = preferred_profile
    build_profile = project.build_profile()

    project_builder.configure(
        project,
        build_profile.builder,
        5
    )

    try:
        app.run(debug=debug, host=host, port=port)

    finally:
        project_builder.stop()



from hathor import __version__

author_name = "BrinkerVII"
author_email = "brinkervii@gmail.com"
package_version = __version__
package_description = "A lua script combobulator aimed at roblox"
package_license = "GPL-3"
package_repository = "https://gitlab.com/rockit-tools/hathor"



import os
import re
import typing
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Union

import yaml
from fs.base import FS
from fs.osfs import OSFS

from hathor.project.information.lua_file import LuaFile
from hathor.project.information.source_root import SourceRoot


@dataclass()
class BuildProfile:
    directory: str = "./build"
    builder: str = "copy"


class Project:
    _root_directory: Path
    _config_path: Path
    _config: Dict = None
    _sources_root: List[SourceRoot]
    _build_profiles: Dict[str, BuildProfile]
    _active_profile: str = "default"

    def __init__(self, config_file: Path):
        self._config_path = config_file
        self._root_directory = self._config_path.parent
        self._sources_root = []

        self._build_profiles = dict()
        self._parse_build_profiles()
        self._scan()

    @property
    def config(self) -> Dict:
        if self._config is None:
            with self._config_path.open("r") as fp:
                self._config = yaml.load(fp, yaml.FullLoader)

        return self._config

    @property
    def build_config(self) -> Dict:
        return self.config.get("build", {})

    def build_profile(self, profile: str = None) -> BuildProfile:
        return self._build_profiles[profile or self._active_profile]

    @property
    def active_profile(self):
        return self._active_profile

    @active_profile.setter
    def active_profile(self, v: str):
        assert v in self._build_profiles, f"{v} is not a valid build profile"
        self._active_profile = v

    @property
    def serve_config(self):
        return self.config.get("serve", {})

    def resolve_user_path(self, path: Union[str, Path]) -> Path:
        if isinstance(path, str):
            # *nix to host os
            if "/" in path:
                path = os.path.sep.join(path.split("/"))

        path = Path(path)
        if not path.is_absolute():
            path = self._root_directory.joinpath(path)

        return path

    def _parse_build_profiles(self):
        profiles = self.build_config.get("profiles", {})

        if "default" in profiles:
            default_profile = profiles["default"]

        else:
            default_profile = {}

            for k, v in typing.get_type_hints(BuildProfile).items():
                default_profile[k] = self.build_config.get(k)

        for profile_name, profile_object in {**profiles, "default": default_profile}.items():
            self._build_profiles[profile_name] = BuildProfile(**profile_object)

    def _scan(self):
        for source_path, properties in self.config["sources"].items():
            if not properties.get("enabled", False):
                continue

            path = self.resolve_user_path(source_path)
            self._sources_root.append(SourceRoot(OSFS(str(path.absolute())), properties))

    def rescan(self):
        self._sources_root = []

        self._scan()

    @property
    def lua_files(self) -> List[LuaFile]:
        files = []

        for root in self._sources_root:
            files.extend(root.lua_files)

        return files

    @property
    def entry_points(self) -> List[LuaFile]:
        return list(filter(lambda f: f.is_entry_point, self.lua_files))

    @property
    def name(self):
        return self.config["name"]

    @property
    def path_compatible_name(self):
        return re.subn(r"\W", "_", self.name.lower())[0]

    @property
    def build_directory(self) -> Path:
        return self.resolve_user_path(self.build_profile().directory)

    @property
    def source_file_systems(self) -> List[FS]:
        return list(map(lambda r: r.fs, self._sources_root))


def find_projects(search_root: Path = None) -> List[Project]:
    search_root = search_root or Path(os.getcwd()).absolute()
    projects = []

    for project_config in search_root.rglob("hathor.project.yml"):
        projects.append(Project(project_config))

    return projects



from typing import Callable, Dict

from fs.base import FS

from hathor.project.builders.copy import build_copy
from hathor.project.builders.modular import build_modular
from hathor.project.information.project import Project

Builder = Callable[[Project, FS], None]

BUILDERS: Dict[str, Builder] = {
    "copy": build_copy,
    "modular": build_modular
}

DEFAULT_BUILDER: str = list(BUILDERS.keys())[0]


class BuilderNotFound(RuntimeError):
    def __init__(self, name: str):
        super().__init__(f"The builder '{name}' could not be found")


def get_builder(name: str) -> Builder:
    if builder := BUILDERS.get(name, None):
        return builder

    else:
        raise BuilderNotFound(name)



import os
import shutil
from pathlib import Path

import click
import requests
import yaml
from click_spinner import spinner
from fs.osfs import OSFS

from hathor import resources
from hathor.project import builders


@click.group()
def cli():
    pass


@cli.command()
@click.option('-p', '--profile', default="default")
def build(profile):
    from hathor.project.information.project import find_projects

    projects = find_projects()
    for project in projects:
        project.active_profile = profile

        build_root = project.build_directory.joinpath(project.path_compatible_name)
        if build_root.exists():
            shutil.rmtree(build_root)

        os.makedirs(build_root, exist_ok=True)

        build_profile = project.build_profile()
        builder = builders.get_builder(build_profile.builder)

        builder(project, OSFS(str(build_root.absolute())))


@cli.command()
def init():
    project_file = Path("hathor.project.yml").absolute()

    if project_file.exists():
        if not click.confirm("The project file will be overwritten if you continue, proceed?"):
            return

    def query_builder():
        b = list(builders.BUILDERS.keys())
        builders_list = "\n".join(b)

        msg = f"Which builder would you like to use for this profile?\n" \
              f"The available builders are:\n\n{builders_list}\n"

        choice = "_"
        while choice not in b:
            click.echo(msg)
            choice = click.prompt("? ", builders.DEFAULT_BUILDER)
            if choice not in b:
                click.echo(f"The builder '{choice}' is not a valid choice!")

        return choice

    def query_build_profile(name: str):
        click.echo(f"Configuring build profile '{name}'")

        profile = {
            "directory": click.prompt("Build directory", "./build"),
            "builder": query_builder()
        }

        return profile

    project = {
        "name": click.prompt("Project name"),
        "metadata": {
            "author": click.prompt("Author"),
            "author_email": click.prompt("Author email")
        },
        "build": query_build_profile("default"),
        "serve": {
            "host": "127.0.01",
            "port": 8080,
        },
        "sources": {
            click.prompt("Default source path", "./src"): {
                "enabled": True
            }
        }
    }

    while click.confirm("Add more sources?"):
        project["sources"][click.prompt("Source directory")] = {
            "enabled": True
        }

    while click.confirm("Add another build profile?"):
        profiles = project["build"].get("profiles", {})
        name = click.prompt("What is the name for this new profile?")

        if not name.strip():
            click.echo("That name is invalid, please try again")
            continue

        profiles[name] = query_build_profile(name)

        project["build"]["profiles"] = profiles

    if "profiles" in project["build"]:
        available_profiles = ["default", *project["build"]["profiles"].keys()]

    else:
        available_profiles = ["default"]

    available_profiles = "\n".join(available_profiles)
    click.echo(f"Currently available build profiles:\n{available_profiles}")

    project["serve"]["build_profile"] = \
        click.prompt("Which build profile would you like to use for serving?", "default")

    with project_file.open("w+") as fp:
        yaml.dump(project, fp, yaml.Dumper, sort_keys=False)

    for source_dir in project["sources"].keys():
        if not source_dir.strip():
            continue

        path = Path(source_dir).absolute()
        os.makedirs(path, exist_ok=True)

    gitignore_path = project_file.parent.joinpath(".gitignore")

    if click.confirm("Add a .gitignore file? This uses the gitignore.io service"):
        ignore_categories = [
            "lua",
            "windows",
            "linux",
            "macos",
            "archive"
        ]

        if click.confirm("Add ignores for IntelliJ?"):
            ignore_categories.append("intellij+all")

        if click.confirm("Add ignores for VSCode?"):
            ignore_categories.append("code")

        click.echo("Downloading gitignore")
        query_string = ",".join(ignore_categories)
        with spinner():
            response = requests.get(f"https://www.toptal.com/developers/gitignore/api/{query_string}")
            if response.status_code > 299:
                click.echo("Gitignore download failed, it will not be appended to the file")
                click.echo(f"{response.request.url}\n{response.status_code} {response.reason}:\n{response.text}")
                response = None

        with gitignore_path.open("w+") as fp:
            fp.write(resources.gitignore_base())

            if response is not None:
                fp.write(response.text)

        click.echo("Gitignore written")

    editorconfig_path = project_file.parent.joinpath(".editorconfig")

    if click.confirm("Add a .editorconfig file?"):
        editorconfig = resources.editorconfig_template({
            "charset": click.prompt("Charset", "utf-8"),
            "end_of_line": click.prompt("End of line", "lf"),
            "indent_size": click.prompt("Indent size", "4"),
            "indent_style": click.prompt("Indent style", "tab"),
            "insert_final_newline": "true" if click.confirm("Insert final newline", True) else "false",
            "max_line_length": click.prompt("Maximum line length", "100"),
            "tab_width": click.prompt("Tab width", "4")
        })

        with editorconfig_path.open("w+") as fp:
            fp.write(editorconfig + "\n")

        click.echo("Editorconfig written")


@cli.command()
def serve():
    from hathor.serve import app
    from hathor.project.information.project import find_projects

    projects = find_projects()
    assert len(projects), f"Only one project can be served at a time, found {len(projects)}"

    app.run(projects[0])


if __name__ == '__main__':
    cli()



from pathlib import Path
from string import Template
from typing import Union, Dict

from cachetools import cached, LRUCache

HERE = Path(__file__).absolute().parent


@cached(cache=LRUCache(64))
def _read_text(path: Union[Path, str]) -> str:
    if isinstance(path, str):
        path = Path(path)

    with HERE.joinpath(path).open("r") as fp:
        return fp.read()


def hathor_loader():
    return _read_text("hathor_loader.lua")


def hathor_stub():
    return _read_text("hathor_stub.lua")


def hathor_run(modules_name: str, entry_point: str) -> str:
    template = Template(_read_text("hathor_run.lua"))

    return template.safe_substitute({
        "modules_name": modules_name,
        "entry_point": entry_point
    })


def gitignore_base():
    return _read_text("gitignore_base.txt")


def editorconfig_template(properties: Dict[str, str]) -> str:
    template = Template(_read_text("editorconfig_template.ini"))

    return template.safe_substitute(properties)



import os

from fs.base import FS

from hathor.project.information.project import Project


def build_copy(project: Project, build_root: FS):
    build_root.makedirs("/", recreate=True)

    for lua_file in project.lua_files:
        build_root.makedirs(os.path.dirname(lua_file.path), recreate=True)

        with build_root.open(lua_file.path, "w+") as fp:
            fp.write(lua_file.read())



import logging
import os
import sys

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
project_path = here
readme_path = os.path.join(project_path, "README.md")

sys.path.insert(0, here)

logging.basicConfig(stream=sys.stderr, level=logging.INFO)


def read_file(path):
    with open(path, "r") as fp:
        return fp.read()


def install_requires():
    requirements = read_file(os.path.join(project_path, "requirements.txt")).split("\n")
    requirements = list(filter(lambda s: not not s, map(lambda s: s.strip(), requirements)))

    return requirements


def main():
    import hathor.__about__ as __about__

    setup(
        name="hathor",
        author=__about__.author_name,
        author_email=__about__.author_email,
        version=__about__.package_version,
        description=__about__.package_description,
        license=__about__.package_license,
        long_description=read_file(readme_path),
        long_description_content_type="text/markdown",
        url=__about__.package_repository,
        classifiers=[
            "Development Status :: 4 - Beta",
            "Programming Language :: Python :: 3.8"
        ],
        keywords=["roblox", "studio", "lua", "build-tool", "build", "tool"],
        packages=find_packages(".", exclude=[
            "tests",
            "tests.*"
        ]),
        package_dir={"": "."},
        include_package_data=True,
        python_requires=">=3.8",
        install_requires=install_requires(),
        entry_points={
            "console_scripts": [
                "hathor=hathor.__main__:cli",
            ]
        }
    )


if __name__ == '__main__':
    main()



import logging
import os
import sys

__version__ = "0.0.10.dev0"

logging.basicConfig(
    stream=sys.stderr,
    level=getattr(logging, os.environ.get("LOG_LEVEL", "INFO"), logging.INFO)
)

LOG = logging.getLogger(__name__)
LOG.info(f"Running version {__version__}")
